/* TIGRE implementation:  KERNEL.C */
/* kernel for reduction engine */
 
/* (C) Copyright 1989   Philip Koopman Jr.  */
/*  Last update: 10/29/89   */
/*  Test code version -- not for distribution  */
 
#include "reduce.h"
#include "heap.h"
#include "sim.h"
 
Celltype *nodea,*nodeb,*nodec ;
 
Valuetype breakval ;
 
/* ______________________________________________________________________ */
Valuetype Evaluate(root,spine_ptr)
   Valuetype root; Celltype **spine_ptr;
  {
    UNIX_REGISTER Valuetype result,resultb ;
    UNIX_REGISTER Valuetype ip;
    register Celltype **temp_spine ;
    UNIX_REGISTER Celltype *parent_ptr, *me_ptr ;
#if DUMP_TRACE || INSTRUMENTS
    Celltype *temp1,*temp2;
    Celltype *temp3;
#else
    UNIX_REGISTER Celltype *temp1,*temp2;
    UNIX_REGISTER Celltype *temp3;
#endif
 
#if DEBUG
    printf(" Evaluate \n");
#endif
    ip = root ;
    temp_spine = spine_ptr ;
 
    do
      {  /* important:  code and data must be in same address space!! */
 
/*       while ( ! (ip.comb & 1) )  /*  --- cures an R2000 cc bug */
         while ( IS_PTR(ip.ptr) )
           {
#if DEBUG_TRACE
     printf("(%X ",temp_spine);
#endif
#if DEBUG
     printf(" push stack element %X\n",ip.ptr);
#endif
              ip.ptr = UNFUDGE(ip.ptr) ;
              *(--temp_spine) = ip.ptr+1 ;   /* push node onto spine stack */
                 sim(wrme,temp_spine);
                 sim(rd,ip.ptr);
              ip = ip.ptr->value ;   /* set ip to child of left node */
#if INSTRUMENTS
    i_counts[DO_UNWIND] ++ ;
#endif
           }
 
      /* now we point to a combinator node */
#if DEBUG_TRACE+1
      if (temp_spine < spine_start) spine_overflow() ;
#endif
#if DEBUG_DUMP
      printf("\n ************evaluate %X***************\n",ip.comb);
      zapflags();  dumpgraph(root_save,0); zapflags();
      printf("\n ******** end of graph dump ***********\n");
#endif
      count += 1 ;
 
#if INSTRUMENTS
    if( (ip.comb>>2) > MAX_TOKEN)
       printf("\nBad ip value = %d",ip.comb);
    else
       i_counts[ip.comb>>2] += 1 ;
#endif
 
      switch(ip.comb >> 2)  /* convert back to normal integer */
      {
 
    /* ______________________________________ */
    case DO_I:
#if DEBUG_TRACE
    printf(" I\n");
#endif
    Use_Me ;
    ip = Rme ;
      sim(rd,Rmex);
    Pop_Spine(1);
    continue;
 
    /* ______________________________________ */
 
    case DO_K:
#if DEBUG_TRACE
    printf(" K\n");
#endif
    Use_Me ;
    ip  = Rme ;
      sim(rd,Rmex);
    Pop_Spine(2) ;
    continue;
 
    /* ______________________________________ */
 
    case DO_S:
#if DEBUG_TRACE
    printf(" S\n");
#endif
    NEWNODE2 ;
    /* non-garbage collection code */
#if INSTRUMENTS
    i_heap += 2 ;
#endif
#if DEBUG
      printf(" (temp1=%X  temp2=%X) ",temp1,temp2);
#endif
    Use_Me ;
    Use_Parent;
    Ltemp1 = ip     = Rme ;
      sim(rd,Rmex);
    Ltemp2          = Rparent ;
      sim(rd,Rparentx);
    Pop_Spine(1);
    Use_Parent;
    Rtemp1 = Rtemp2 = Rparent ;
      sim(rd,Rparentx);
      dsim(wr,Ltemp2x);
      dsim(wr,Ltemp1x);
    Rparent.child = FUDGE(temp2) ;
    Lparent.child = FUDGE(temp1) ;
      dsim(wr,Lparentx);
    *(temp_spine) = temp1+1 ;
      sim(wrme,temp_spine);
    continue ;
 
    /* ______________________________________ */
 
    case DO_SPRIME:
#if DEBUG_TRACE
    printf(" S'\n");
#endif
    NEWNODE3 ;
#if INSTRUMENTS
    i_heap += 3 ;
#endif
#if DEBUG
    printf(" (temp1=%X  temp2=%X  temp3=%X) ",temp1,temp2,temp3);
#endif
    Use_Me ;
    Use_Parent;
    Ltemp1  = ip             = Rme ;
      sim(rd,Rmex);
    Rtemp1.child  = FUDGE(temp2) ;
      dsim(wr,Ltemp1x);
    Ltemp2        = Rparent ;
      sim(rd,Rparentx);
    Pop_Spine(2) ;
    Use_Me ;
    Use_Parent;
    Ltemp3 = Rme ;
      sim(rd,Rmex);
    Lparent.child = FUDGE(temp1) ;
    Rtemp2  = Rtemp3  = Rparent ;
      sim(rd,Rparentx);
      dsim(wr,Ltemp2x);
      dsim(wr,Ltemp3x);
    Rparent.child = FUDGE(temp3) ;
      dsim(wr,Lparentx);
    *(temp_spine) = temp1+1;
      sim(wrme,temp_spine);
    continue;
 
    /* ______________________________________ */
 
    case DO_B:
#if DEBUG_TRACE
    printf(" B\n");
#endif
    NEWNODE1;
#if INSTRUMENTS
    i_heap += 1 ;
#endif
#if DEBUG
    printf(" (temp1=%X) ",temp1);
#endif
    Use_Me ;
    Use_Parent;
    ip           = Rme ;
      sim(rd,Rmex);
    Ltemp1       = Rparent  ;
      sim(rd,Rparentx);
    Pop_Spine(2) ;
    Use_Me ;
    Lme = ip ;
    Rtemp1       = Rme ;
      sim(rd,Rmex);
      dsim(wr,Ltemp1x);
    Rme.child = FUDGE(temp1) ;
      dsim(wr,Lmex);
    continue;
 
    /* ______________________________________ */
 
    case DO_BSTAR:
#if DEBUG_TRACE
    printf(" B*\n");
#endif
    NEWNODE2 ;
#if INSTRUMENTS
    i_heap += 2 ;
#endif
#if DEBUG
    printf(" (temp1=%X  temp2=%X) ",temp1,temp2);
#endif
    Use_Me ;
    Use_Parent;
    ip           = Rme  ;
      sim(rd,Rmex);
    Ltemp1       = Rparent ;
      sim(rd,Rparentx);
    Rtemp1.child = FUDGE(temp2) ;
      dsim(wr,Ltemp1x);
    Pop_Spine(2) ;
    Use_Me ;
    Use_Parent;
    Ltemp2       = Rme ;
      sim(rd,Rmex);
    Rtemp2       = Rparent ;
      sim(rd,Rparentx);
      dsim(wr,Ltemp2x);
    Rparent.child = FUDGE(temp1) ;
    Lparent      = ip ;
      dsim(wr,Lparentx);
    Pop_Spine(1) ;
    continue;
 
    /* ______________________________________ */
 
    case DO_C:
#if DEBUG_TRACE
    printf(" C\n");
#endif
    NEWNODE1;
#if INSTRUMENTS
    i_heap += 1 ;
#endif
#if DEBUG
    printf(" (temp1=%X) ",temp1);
#endif
    Use_Me ;
    Ltemp1        = ip              = Rme  ;
      sim(rd,Rmex);
    Pop_Spine(1);
    Use_Parent;
    Use_Me ;
    Rtemp1        = Rparent ;
      sim(rd,Rparentx);
      dsim(wr,Ltemp1x);
    Rparent       = Rme ;
      sim(rd,Rmex);
    Lparent.child  = FUDGE(temp1) ;
      dsim(wr,Lparentx);
    *(temp_spine) = temp1+1 ;
      sim(wrme,temp_spine);
    continue ;
 
    /* ______________________________________ */
 
    case DO_CPRIME:
#if DEBUG_TRACE
    printf(" C'\n");
#endif
    NEWNODE2;
#if INSTRUMENTS
    i_heap += 2 ;
#endif
#if DEBUG
    printf(" (temp1=%X  temp2=%X) ",temp1,temp2);
#endif
    Use_Me ;
    Use_Parent;
    Ltemp1       =  ip           = Rme  ;
      sim(rd,Rmex);
    Rtemp1.child = FUDGE(temp2) ;
      dsim(wr,Ltemp1x);
    Ltemp2       = Rparent ;
      sim(rd,Rparentx);
    Pop_Spine(2) ;
    Use_Me ;
    Use_Parent;
    Rtemp2       = Rparent ;
      sim(rd,Rparentx);
      dsim(wr,Ltemp2x);
    Rparent       = Rme ;
      sim(rd,Rmex);
    Lparent.child = FUDGE(temp1) ;
      dsim(wr,Lparentx);
    *(temp_spine) = temp1+1 ;
      sim(wrme,temp_spine);
    continue;
 
    /* ______________________________________ */
 
    case DO_IF:
#if DEBUG_TRACE
    printf(" IF\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine);
    Pop_Spine(1);
    Use_Me ;
    Use_Parent;
    if (result.literal)
      {
        ip = Rparent = Rme ;
          sim(rd,Rmex);
        Lparent.comb = do_i_value.value.comb ;
          dsim(wr,Lparentx);
        Pop_Spine(2);
        continue ;
      }
    ip = Rparent ;
    Lparent.comb = do_i_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2);
    continue ;
 
    /* ______________________________________ */
 
    case DO_LIT:
    Use_Me ;
#if DEBUG_TRACE
    printf(" LIT(%d)",Rme.literal);
#endif
    result = Rme ;
      sim(rd,Rmex);
    Pop_Spine(1) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_ZERO:
#if DEBUG_TRACE
    printf(" 0\n");
#endif
    result.literal = 0;
    break ;
 
    /* ______________________________________ */
 
    case DO_TRUE:
#if DEBUG_TRACE
    printf(" TRUE\n");
#endif
    result.literal = TRUE ;
    break ;
 
    /* ______________________________________ */
 
    case DO_FALSE:
#if DEBUG_TRACE
    printf(" FALSE\n");
#endif
    result.literal = FALSE;
    break ;
 
    /* ______________________________________ */
 
    case DO_NIL:
#if DEBUG_TRACE
    printf(" NIL\n");
#endif
    result.literal = -131313 ;
    break ;
 
    /* ______________________________________ */
 
    case DO_NL:
#if DEBUG_TRACE
    printf(" NL\n");
#endif
    printf("\n");
    result.literal = NULL ;
    break;
 
    /* ______________________________________ */
 
    case DO_NOT:
#if DEBUG_TRACE
    printf(" NOT\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result  =  Evaluate(Rme,temp_spine) ;
    Use_Me ;
    result.literal = ! result.literal ;
    Rme = result ;
    Lme.comb    = do_lit_value.value.comb ;
      dsim(wr,Lmex);
    Pop_Spine(1) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_ABS:
#if DEBUG_TRACE
    printf(" ABS\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result  = Evaluate(Rme,temp_spine) ;
    Use_Me ;
    if (result.literal < 0 )  result.literal = 0 - result.literal ;
    Rme = result ;
    Lme.comb    = do_lit_value.value.comb ;
      dsim(wr,Lmex);
    Pop_Spine(1) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_PRINT:
#if DEBUG_TRACE
    printf(" PRINT\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result  = Evaluate(Rme,temp_spine) ;
    Use_Me ;
    Rme = result ;
    Lme.comb    = do_lit_value.value.comb ;
      dsim(wr,Lmex);
    Pop_Spine(1) ;
#if DEBUG_TRACE
    printf("\n---->\n");
#endif
    printf(" %d",result.literal);
#if DEBUG_TRACE
    printf("\n\n");
#endif
    break ;
 
    /* ______________________________________ */
 
    case DO_PLUS:
#if DEBUG_TRACE
    printf(" +\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal + resultb.literal ;
    Rparent = result ;
    Lparent.comb  = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_MINUS:
#if DEBUG_TRACE
    printf(" -\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal - resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_ONE:
#if DEBUG_TRACE
    printf(" 1\n");
#endif
    result.literal = 1 ;
    break ;
 
    /* ______________________________________ */
 
    case DO_TWO:
#if DEBUG_TRACE
    printf(" 2\n");
#endif
    result.literal = 2 ;
    break ;
 
    /* ______________________________________ */
 
    case DO_THREE:
#if DEBUG_TRACE
    printf(" 3\n");
#endif
    result.literal = 3 ;
    break ;
 
    /* ______________________________________ */
 
    case DO_TIMES:
#if DEBUG_TRACE
    printf(" *\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal * resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_MOD:
#if DEBUG_TRACE
    printf(" MOD\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal % resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_DIV:
#if DEBUG_TRACE
    printf(" DIV\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal / resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
    /* ______________________________________ */
 
    case DO_FLOOR:
#if DEBUG_TRACE
    printf(" FLOOR\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    if (    ( ( result.literal ^ resultb.literal ) < 0 )
         && (result.literal % resultb.literal          )  )
     {  result.literal = result.literal / resultb.literal ;
        result.literal-- ;
     }
    else
     {  result.literal = result.literal / resultb.literal ;
     }
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_LESS:
#if DEBUG_TRACE
    printf(" <\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal < resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_GREAT:
#if DEBUG_TRACE
    printf(" >\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal > resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_EQ:
#if DEBUG_TRACE
    printf(" =\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    Use_Parent;
      sim(rd,Rparentx);
    resultb = Evaluate(Rparent,temp_spine);
    Use_Parent;
    result.literal = result.literal == resultb.literal ;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_AND:
#if DEBUG_TRACE
    printf(" AND\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    if(result.literal)
      { Use_Parent;
           sim(rd,Rparentx);
        resultb = Evaluate(Rparent,temp_spine);
        if (resultb.literal) result.literal = TRUE ;
        else    result.literal = FALSE ;
      }
      else    result.literal = FALSE ;
    Use_Parent;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_OR:
#if DEBUG_TRACE
    printf(" OR\n");
#endif
    Use_Me ;
      sim(rd,Rmex);
    result = Evaluate(Rme,temp_spine) ;
    if(!result.literal)
      { Use_Parent;
          sim(rd,Rparentx);
        resultb = Evaluate(Rparent,temp_spine);
        if (!resultb.literal) result.literal = FALSE ;
        else    result.literal = TRUE ;
      }
      else    result.literal = TRUE ;
    Use_Parent;
    Rparent = result ;
    Lparent.comb = do_lit_value.value.comb ;
      dsim(wr,Lparentx);
    Pop_Spine(2) ;
    break ;
 
    /* ______________________________________ */
 
    case DO_P:
#if DEBUG_TRACE
    printf(" P");
#endif
    /*  P returns the address of the parent node RHS as its
     *  result.  Also, all I-nodes between the parent and
     *  me nodes are shorted by re-writing Lparent.
     * 1) comparisons can compare on this address.
     * 2) The U operator can then look at the result & LHS
     */
    Use_Parent;
    Use_Me ;
    result.child = parent_ptr - 1 ;
    Lparent.child = FUDGE(me_ptr - 1);
      sim(wr,Lparentx);
    Pop_Spine(2);
    break ;
 
    /* ______________________________________ */
 
    case DO_U:
#if DEBUG_TRACE
    printf(" U");
#endif
    /* Evaluate Rparent subtree.  That subtree's P combinator
     * will return pointer to RHS of node, and guarantee that
     * the LHS of that node points to the node whose RHS
     * contains the other pair parameter.
     */
    Use_Parent;
    /* use temp2 as a scratch pointer to the P subtree */
    /* apparent bug in TURBOC won't allow   temp2 = Evaluate(...).ptr ; */
      sim(rd,Rparentx);
    result = Evaluate(Rparent,temp_spine) ;
    NEWNODE1; 
 
#if INSTRUMENTS
    i_heap += 1 ;
#endif
#if DEBUG
    printf(" (temp1=%X) ",temp1);
#endif
    temp2 = result.ptr ;
    Use_Parent ;
    Use_Me ;
    Lparent.child = FUDGE(temp1) ;
    Rparent = Rtemp2 ;
      dsim(rd,Ltemp2x);
      dsim(wr,Lparentx);
    Ltemp1 = ip     = Rme ;
      sim(rd,Rmex);
    Rtemp1          = (UNFUDGE(Ltemp2.child)+1)->value ;
      sim(rd,(UNFUDGE(Ltemp2.child)+1) );
      dsim(wr,Ltemp1x);
    *(temp_spine) = temp1+1 ;
      sim(wrme,temp_spine);
    continue ;
 
 
    /* ______________________________________ */
 
   case SPINE_UNDERFLOW: printf("\n\n spine stack underflow\n");
      exit(-1);
 
 
/* Supercombinator definition for fibonacci */
 
#define fast_mapval(x)  ((x * 4) + 1)
/* ______________________________________________________________________ */
 
case DO_S3:
#if DEBUG_TRACE
    printf(" $3\n");
#endif
  Use_Me;
    sim(rd,Rmex);
  result = Evaluate(Rme,temp_spine) ;
  if (result.literal < 3 )
   { result.literal = 1 ;
     Pop_Spine(1);
     break ;
   }
 
  NEWNODEN(6);
  Use_Me;
 
  resultb = Rme ;
    sim(rd,Rmex);
 
  (temp1+0)->value.comb = fast_mapval(DO_LIT);
  (temp1+1)->value.literal = result.literal - 1 ;
    dsim(wr,temp1+0);
 
  (temp1+2)->value.comb = fast_mapval(DO_S3);
  (temp1+3)->value.child = FUDGE(temp1+0);
    dsim(wr,temp1+2);
 
  (temp1+4)->value.comb = fast_mapval(DO_PLUS);
  (temp1+5)->value.child = FUDGE(temp1+2);
    dsim(wr,temp1+4);
 
  (temp1+6)->value.comb = fast_mapval(DO_LIT);
  (temp1+7)->value.literal = result.literal - 2 ;
    dsim(wr,temp1+6);
 
  (temp1+8)->value.comb = fast_mapval(DO_S3);
  (temp1+9)->value.child = FUDGE(temp1+6);
    dsim(wr,temp1+8);
 
  (temp1+10)->value.child = FUDGE(temp1+4);
  (temp1+11)->value.child = FUDGE(temp1+8);
    dsim(wr,temp1+10);
 
  Lme.comb = ip.comb = fast_mapval(DO_I);
  Rme.child = FUDGE(temp1+10);
    dsim(wr,Lmex);
 
/*  weird!!! */
  continue;
/* ______________________________________________________________________ */
 
case DO_S6:
#if DEBUG_TRACE
    printf(" $6\n");
#endif
  Use_Me;
    sim(rd,Rmex);
  result = Evaluate(Rme,temp_spine) ;
  if (result.literal < 2 )
   { result.literal = 1 ;
     Pop_Spine(1);
     break ;
   }
  NEWNODEN(8);
 
  Use_Me;
  resultb = Rme ;
    sim(rd,Rmex);
 
  (temp1+0)->value.comb = fast_mapval(DO_LIT);
  (temp1+1)->value.literal = result.literal - 1 ;
    dsim(wr,temp1+0);
 
  (temp1+2)->value.comb = fast_mapval(DO_S6);
  (temp1+3)->value.child = FUDGE(temp1+0);
    dsim(wr,temp1+2);
 
  (temp1+4)->value.comb = fast_mapval(DO_PLUS);
  (temp1+5)->value.child = FUDGE(temp1+2);
    dsim(wr,temp1+4);
 
  (temp1+6)->value.comb = fast_mapval(DO_LIT);
  (temp1+7)->value.literal = result.literal - 2 ;
    dsim(wr,temp1+6);
 
  (temp1+8)->value.comb = fast_mapval(DO_S6);
  (temp1+9)->value.child = FUDGE(temp1+6);
    dsim(wr,temp1+8);
 
  (temp1+10)->value.child = FUDGE(temp1+4);
  (temp1+11)->value.child = FUDGE(temp1+8);
    dsim(wr,temp1+10);
 
  (temp1+12)->value.comb = fast_mapval(DO_PLUS);
  (temp1+13)->value.comb = fast_mapval(DO_ONE);
    dsim(wr,temp1+12);
 
  (temp1+14)->value.child = FUDGE(temp1+12);
  (temp1+15)->value.child = FUDGE(temp1+10);
    dsim(wr,temp1+14);
 
  Lme.comb= ip.comb = fast_mapval(DO_I);
  Rme.child = FUDGE(temp1+14);
    dsim(wr,Lmex);
  continue;
/* ______________________________________________________________________ */
 
 
case DO_S1:
case DO_S2:
case DO_S4:
case DO_S5:
case DO_S7:
   case DO_S8:
/* note: s9 changed to /_  */
   case DO_S10:
   case DO_S11:
   case DO_S12:
   case DO_S13:
   case DO_S14:
   case DO_S15:
   case DO_S16:
   case DO_S17:
   case DO_S18:
   case DO_S19:
   default: printf("\n\n jump to bad combinator\n");
      exit(-1);
   }
#if DEBUG_TRACE
     printf(" %X)",temp_spine);
#endif
    return(result);
  }
    while ( TRUE ) ;
   /* unreachable code    ---  return(result);  */
  }
 
/* _____________________________________________________________ */
 
INT_32 mapval(toknum)
  int toknum ;
  { int temp_toknum ;
    temp_toknum = toknum << 2 ;  /* force to odd value */
    temp_toknum += 1 ;
    return(temp_toknum);}
/*  this translates token integers to execution addresses */
 
