/* TIGRE Graph Reducer -- Master header file  REDUCE.H */
/* (C) Copyright 1989   Philip Koopman Jr.  */
/*  Last update:  4/17/89   */
/*  Test code version -- not for distribution  */
 
#define TRUE   1
#define FALSE  0
/* The TURBOC timing routines are primitive -- so don't bother to use them */
 
#ifdef TURBOC
#define UNIX FALSE
#else
#define UNIX TRUE
#endif
 
#ifdef VAX          /* adapt to Vax assembler kernel */
#define VAX_ASM TRUE
#else
#define VAX_ASM FALSE
#endif
 
#ifdef INST
#define INSTRUMENTS TRUE
#else
#define INSTRUMENTS FALSE
#endif
 
#ifdef TRACE
#define DUMP_TRACE TRUE
#else
#define DUMP_TRACE FALSE
#endif
 
 
#define DEBUG  FALSE
#define DEBUG_TRACE FALSE+DEBUG
#define DEBUG_HEAP  DEBUG+FALSE
#define DEBUG_DUMP  FALSE
 
/* If not UNIX, then operating in TURBOC */
 
#include <stdio.h>
 
#if UNIX
#include <sys/types.h>
#include <sys/times.h>
/* time measurement definitions */
#define HZ      60
#define INT_32  int
 
#else   /* TURBOC time telling, etc. */
#include <time.h>
#include <stdlib.h>
#define INT_32  long
#define HZ      CLK_TCK
#endif
 
#define FIRST_PASS_FLAG FALSE   /* whether input gets echoed */
 
#if UNIX
#define HEAPSIZE    100000               /* # nodes in heap */
#define SPINESIZE   100000               /* # cells in spine stack */
#define UNIX_REGISTER  register
#else
#define HEAPSIZE    4000               /* # nodes in heap */
#define SPINESIZE   800               /* # cells in spine stack */
#define UNIX_REGISTER /* no-op */
#endif
 
#define  MAX_TOKEN 255
#define PROGSIZE    HEAPSIZE/2         /* max size of initial program */
 
#if VAX_ASM
#define HEAP_STRIDE 3
#else
#define HEAP_STRIDE 2
#endif
 
typedef union Valuetype {
     struct Celltype *child ;          /* child in combinator tree */
     struct Celltype *ptr ;            /* pointer when used in as intr ptr */
     INT_32 comb ;
     INT_32 literal  ;
     }  Valuetype ;
 
typedef struct Celltype {
     Valuetype value ;
     }  Celltype ;
 
extern Valuetype Evaluate();
extern void exit();
 
extern void dumpgraph();
extern void spine_under();
extern void spine_overflow();
 
extern void collect_garbage();
extern void fixup_stack();
extern void init_heap();  /* init heap memory -- leaves first xx nodes free */
extern void zapflags();  /* set all flags in heap arrays to false */
extern INT_32 mapval();
extern Celltype *newnode();
extern Celltype *newnode_continue();
extern void newnode2();
extern void newnode2_continue();
extern void newnode3();
extern void newnode3_continue();

 
extern Celltype do_lit_value,do_i_value ; /* save DO_LIT and DO_I values */
extern int error_flag ;
extern Celltype *temp_pointer ;
 
extern Valuetype ip ;
extern Celltype *root_save ;              /* root of evalution tree */
extern Celltype *next_free_node ;         /* heap free list pointer */
extern Celltype *heap_start, *heap_end ;  /* boundaries of in-use heap */
extern Celltype *from_heap, *to_heap ;    /* two heap buffers */
extern Celltype **spine_start,**spine_end ;  /* boundaries of spine memory */
extern Celltype **spine_stack ; /* points to topmost element on spine stack */
extern Celltype *spine_mem[SPINESIZE+100];    /* spine stack storage */
extern Celltype **fixup_start ;
 
extern long count;   /* counts number of combinators used */
extern int gc ;      /* counts number of garbage collections used */
extern int error_flag ;  /* flag if graph has an error */
 
extern long i_counts[MAX_TOKEN+10]; /* combinator freq. counts */
extern long i_heap   ;  /* number of heap nodes allocated */
 
/* Use this macro before first reference to me node */
#define  Use_Me      me_ptr = *temp_spine ;  \
                      sim(rdme,temp_spine);
/* Rme points to right hand side of current node */
#define  Rme       (me_ptr)->value
#define  Lme       (me_ptr-1)->value
#define  Rmex      (me_ptr)
#define  Lmex      (me_ptr-1)
 
/* Use this macro before first reference to parent node */
#define  Use_Parent  parent_ptr = *(temp_spine + 1) ; \
                      sim(rdp,(temp_spine+1) );
/* Rparent points to right hand side of parent node */
#define  Rparent     (parent_ptr)->value
#define  Lparent     (parent_ptr-1)->value
#define  Rparentx    (parent_ptr)
#define  Lparentx    (parent_ptr-1)
 
/* Use this macro before first reference to grandparent node */
#define  Use_Grandparent  /* nop */              \
                      sim(rdg,(temp_spine+2) );
/* Rparent points to right hand side of grandparent node */
#define  Rgrandparent     (*(temp_spine + 2))->value
#define  Lgrandparent     ((*(temp_spine + 2))-1)->value
#define  Rgrandparentx    (*(temp_spine + 2))
#define  Lgrandparentx    ((*(temp_spine + 2))-1)
 
/* Use this macro before first reference to greatgrandparent node */
#define  Use_Greatgrandparent  /* nop */
/* Rparent points to right hand side of greatgrandparent node */
#define  Rgreatgrandparent     (*(temp_spine + 3))->value
#define  Lgreatgrandparent     ((*(temp_spine + 3))-1)->value
#define  Rgreatgrandparentx     (*(temp_spine + 3))
#define  Lgreatgrandparentx     ((*(temp_spine + 3))-1)
 
/* Rtemp1 points to right hand side of 1st temp node */
#define  Ltemp1    temp1->value
#define  Rtemp1    (temp1+1)->value
/* Rtemp2 points to right hand side of 2nd temp node */
#define  Ltemp2    temp2->value
#define  Rtemp2    (temp2+1)->value
/* Rtemp3 points to right hand side of 3rd temp node */
#define  Ltemp3    temp3->value
#define  Rtemp3    (temp3+1)->value
 
#define  Ltemp1x    temp1
#define  Rtemp1x    (temp1+1)
#define  Ltemp2x    temp2
#define  Rtemp2x    (temp2+1)
#define  Ltemp3x    temp3
#define  Rtemp3x    (temp3+1)
 
#define Pop_Spine(x)  temp_spine += x
 
/* token definitions for combinators */
#define  DO_I     1
#define  DO_K     2
#define  DO_S     3
#define  DO_SPRIME 4
#define  DO_B     5
#define  DO_C     6
#define  DO_IF    7
#define  DO_LIT   8
#define  DO_PLUS  9
#define  DO_MINUS 10
#define  DO_LESS  11
#define  DO_TIMES 12
#define  DO_ONE   13
#define  DO_ZERO  14
#define  DO_EQ    15
#define  DO_GREAT 16
#define  DO_P     17
#define  DO_U     18
/* #define  DO_??    19 */
#define  DO_NOT   20
#define  DO_ABS   21
#define  DO_TRUE  22
#define  DO_FALSE 23
#define  DO_NIL   24
#define  DO_PRINT 25
#define  DO_NL    26
#define  DO_AND   27
#define  DO_OR    28
#define  DO_MOD   29
#define  DO_BSTAR 30
#define  DO_CPRIME 31
#define  DO_TWO   32
#define  DO_THREE 33
#define  DO_DIV   34
#define  SPINE_UNDERFLOW 35   /* spine stack pointer for stack underflow */
 
#define  DO_S1    36
#define  DO_S2    37
#define  DO_S3    38
#define  DO_S4    39
#define  DO_S5    40
#define  DO_S6    41
#define  DO_S7    42
#define  DO_S8    43
#define  DO_FLOOR    44
#define  DO_S10   45
#define  DO_S11   46
#define  DO_S12   47
#define  DO_S13   48
#define  DO_S14   49
#define  DO_S15   50
#define  DO_S16   51
#define  DO_S17   52
#define  DO_S18   53
#define  DO_S19   54
 
#define  NOT_A_TOKEN 1001
#define  CYCLE       1002
#define  LPAREN      1003
#define  RPAREN      1004
#define  EOF_TOKEN   1005
#define  KEYWORD     1006
#define  DO_NODE     1007
#define  DO_UNWIND   0
 
 
#if VAX_ASM
/*   jsb occupies 2 bytes before the lhs when fudged*/
#define FUDGE(x)        ( (Celltype *) ( ((INT_32) (x-1)) |  2))
#define UNFUDGE(x)      ( (Celltype *) ( ((INT_32) (x+1)) & -3))
/* forwarded cell has bit 1 cleared */
#define MAKE_FORWARD(x) ( (Celltype *) ( ((INT_32) (x  )) & -3))
#define UNFORWARD(x)    ( (Celltype *) ( ((INT_32) (x  )) |  2))
 
#define IS_PTR(x)        ( ! ( ((INT_32) (x)) & 1) )
#define IS_COMB(x)           ( ((INT_32) (x)) & 1)
#define IS_FORWARDED(x)  ( ! ( ((INT_32) (x)) & 3) )
 
#else
 
#define FUDGE(x)     x
#define UNFUDGE(x)   x
#define MAKE_FORWARD(x) ( (Celltype *) ( ((INT_32) (x)) |  2))
#define UNFORWARD(x)    ( (Celltype *) ( ((INT_32) (x)) & -3))
#define IS_PTR(x)        ( ! ( ((INT_32) (x)) & 0x80000001) )
#define IS_COMB(x)           ( ((INT_32) (x)) & 0x80000001)
#define IS_FORWARDED(x)      ( ((INT_32) (x)) & 2)
 
#endif
 
 
#if UNIX
/* UNMARKED corresponds to a subroutine call instruction jsb immediate long */
#define UNMARKED 0x9F160101
/* note: 01 is a VAX no-op */
#else
/* need a 16-bit quantity for the PC */
#define UNMARKED    1234
#endif
#define MARKED      1
/* #define MULTIMARKED 2 */
 
